/*
								+--------------------------------+
								|                                |
								|   ***  OpenCL utilities  ***   |
								|                                |
								|  Copyright  -tHE SWINe- 2010  |
								|                                |
								|          ClUtils.cpp           |
								|                                |
								+--------------------------------+
*/

#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#define _ASSERTE(x) assert(x)
#include <algorithm>
#if defined(_WIN32) || defined(_WIN64)
#define WIN32_LEAN_AND_MEAN // don't include unnecessary stuff from windows.h
#define NOMINMAX // don't define min and max macros, we prefer to use std::min() and std::max()
#include <windows.h> // Sleep()
#include <io.h> //_access()
#define access(f,m) _access(f,m)
#else //_WIN32 || _WIN64
#include <unistd.h> // sleep(), access()
#endif //_WIN32 || _WIN64
#include "Integer.h"
#include "Hash.h"
#include <CL/opencl.h>
#include "ClUtils.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// MSVC 'for' scoping hack

/*
 *								=== CCLUtils ===
 */

int CCLUtils::n_OpenCL_Init(cl_context *p_context, int n_device_type, bool b_implementation_profile_selection)
{
	cl_context_properties p_props[3] = {cl_context_properties(CL_CONTEXT_PLATFORM), 0, 0};
	if(!b_implementation_profile_selection) {
		cl_uint n_platform_num = 0;
		int n_result;
		if((n_result = clGetPlatformIDs(0, NULL, &n_platform_num)) != CL_SUCCESS)
			return n_result; // error
		if(n_platform_num == 0)
			return CL_DEVICE_NOT_FOUND; // no OpenCL capable platform available
		// get number of OpenCL platforms

		std::vector<cl_platform_id> platform_list;
		try {
			platform_list.resize(n_platform_num);
		} catch(std::bad_alloc&) {
			return CL_OUT_OF_HOST_MEMORY;
		}
		_ASSERTE(platform_list.size() <= /*CL*/UINT_MAX);
		if((n_result = clGetPlatformIDs(cl_uint(platform_list.size()),
		   &platform_list[0], &n_platform_num)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(platform_list.size() == n_platform_num);
		// get list of platform id's

		size_t n_full_platform = 0; // if no full platform is found, just choose the first one
		for(size_t i = 0, n = platform_list.size(); i < n; ++ i) {
			char p_s_profile_name[20] = {0};
			clGetPlatformInfo(platform_list[i], CL_PLATFORM_PROFILE, sizeof(p_s_profile_name), p_s_profile_name, NULL);
			p_s_profile_name[sizeof(p_s_profile_name) / sizeof(p_s_profile_name[0]) - 1] = 0; // make sure it's terminated

			if(!strcmp(p_s_profile_name, "FULL_PROFILE")) {
				n_full_platform = i;
				break;
			}
		}
		// find "FULL_PROFILE" platform

		p_props[1] = cl_context_properties(platform_list[n_full_platform]);
	}
	// get first available "FULL_PROFILE" OpenCL platform

	cl_int n_result;
	cl_context h_context = clCreateContextFromType((b_implementation_profile_selection)?
		NULL : p_props, n_device_type, 0, 0, &n_result);
	if(n_result != CL_SUCCESS)
		return n_result;
	// create context

	*p_context = h_context;

	return CL_SUCCESS;
}

int CCLUtils::n_GetDeviceList(cl_context h_context, std::vector<cl_device_id> &r_device_list)
{
	{
		/** @todo - if CL 1.1, it is possible to use CL_CONTEXT_NUM_DEVICES */
		size_t n_device_size;
		int n_result;
		if((n_result = clGetContextInfo(h_context, CL_CONTEXT_DEVICES, 0, 0, &n_device_size)) != CL_SUCCESS)
			return n_result;
		size_t n_device_num = n_device_size / sizeof(cl_device_id); // it's in bytes
		if(!n_device_num)
			return CL_DEVICE_NOT_FOUND; // no devices
		try {
			r_device_list.resize(n_device_num);
		} catch(std::bad_alloc&) {
			return CL_OUT_OF_HOST_MEMORY;
		}
		if((n_result = clGetContextInfo(h_context, CL_CONTEXT_DEVICES, n_device_size, &r_device_list[0], 0)) != CL_SUCCESS)
			return n_result;
	}
	// get list of devices

	return CL_SUCCESS;
}

size_t CCLUtils::n_Get_MaxGFlops_DeviceId(cl_device_id *p_device_id, cl_context h_context)
{
	std::vector<cl_device_id> device_list;
	if(n_GetDeviceList(h_context, device_list) != CL_SUCCESS)
		return -1;
	// get all the devices

	size_t n_best_device = 0;
	{
		double f_best_gflops = -1;
		for(size_t i = 0, n = device_list.size(); i < n; ++ i) {
			cl_uint n_speed, n_processors;
			if(clGetDeviceInfo(device_list[i], CL_DEVICE_MAX_CLOCK_FREQUENCY,
			   sizeof(n_speed), &n_speed, NULL) != CL_SUCCESS ||
			   clGetDeviceInfo(device_list[i], CL_DEVICE_MAX_COMPUTE_UNITS,
			   sizeof(n_processors), &n_processors, NULL) != CL_SUCCESS)
				return -1;
			double f_gflops = n_speed * n_processors;
			if(!i || f_best_gflops < f_gflops) {
				f_best_gflops = f_gflops;
				n_best_device = i;
			}
		}
	}
	// go trough devices, and estimate computing power (not real gflops)

	*p_device_id = device_list[n_best_device];
	// write handle to the best device

	return n_best_device;
}

size_t CCLUtils::n_Get_MaxGFlops_DeviceId(cl_context h_context)
{
	cl_device_id h_dummy;
	return n_Get_MaxGFlops_DeviceId(&h_dummy, h_context);
}

/*
 *								=== ~CCLUtils ===
 */

/*
 *								=== CCLDeviceParams ===
 */

int CCLDeviceParams::n_GetDeviceInfoString(std::string &r_s_str, int n_name)
{
	return n_GetDeviceInfoString(r_s_str, m_h_device, n_name);
}

int CCLDeviceParams::n_GetDeviceInfoString(std::string &r_s_str, cl_device_id h_device, int n_name)
{
	size_t n_name_size;
	int n_result;
	if((n_result = clGetDeviceInfo(h_device, n_name, 0, NULL, &n_name_size)) != CL_SUCCESS)
		return n_result;
	// get string length

	try {
		r_s_str.resize(n_name_size + 1);
		// allocate string

		if((n_result = clGetDeviceInfo(h_device, n_name, n_name_size, &r_s_str[0], NULL)) != CL_SUCCESS)
			return n_result;
		r_s_str[n_name_size] = 0;
		r_s_str.resize(strlen(r_s_str.c_str()));
		// get string
	} catch(std::bad_alloc&) {
		return CL_OUT_OF_HOST_MEMORY;
	}

	return CL_SUCCESS;
}

bool CCLDeviceParams::b_IsExtensionSupported(const char *p_s_extension_name)
{
	return b_IsExtensionSupported(m_h_device, p_s_extension_name);
}

bool CCLDeviceParams::b_IsExtensionSupported(cl_device_id h_device, const char *p_s_extension_name)
{
	std::string s_exts;
	if(n_GetDeviceInfoString(s_exts, h_device, CL_DEVICE_EXTENSIONS) != CL_SUCCESS)
		return false;
	// get device extension string

	size_t n_ext_name_length = strlen(p_s_extension_name);
	size_t n_pos = 0, n_end_pos = s_exts.length() - n_ext_name_length;
	while((n_pos = s_exts.find(p_s_extension_name, n_pos)) != std::string::npos) {
		if((!n_pos || isspace(s_exts[n_pos - 1])) &&
		   (n_pos >= n_end_pos || isspace(s_exts[n_pos + n_ext_name_length])))
			return true;
		// there must be space before and after it
	}
	// attempt to find the extension

	return false;
	// extension not found
}

/**
 *	@def PRIsizeB
 *
 *	@brief simple macro for printing file sizes in user-readable format
 *
 *	This macro is used in conjunction with PRIsizeBparams.
 */
#define PRIsizeB "%.*f %s"

/**
 *	@def PRIsizeBparams
 *
 *	@brief simple macro for expanding size in bytes to number of fractional
 *		digits, value and unit
 *
 *	This is used with PRIsizeB for printing sizes. Usage is as follows:
 *
 *@code
 *	void Test()
 *	{
 *		printf("%d = " PRIsizeB "B\n", 256, PRIsizeBparams(256));
 *		printf("%d = " PRIsizeB "B\n", 1024 + 256, PRIsizeBparams(1024 + 256));
 *		printf("%d = " PRIsizeB "B\n", 1048576, PRIsizeBparams(1048576));
 *	}
 *
 *	output:
 *
 *	256 = 256 B
 *	1280 = 1.25 kB
 *	1048576 = 1.00 MB
 *@endcode
 *
 *	More extended version of this macro is PRIsizeBparamsExt.
 *
 *	@param[in] s is size in bytes
 */
#define PRIsizeBparams(s) (((s) < 1024)? 0 : 2), (((s) < 1024)? (s) : ((s) < 1048576)? \
	(s) / 1024.0 : ((s) < 1073741824)? (s) / 1048576.0 : ((s) < 1099511627776)? \
	(s) / 1073741824.0 : (s) / 1099511627776.0), (((s) < 1024)? "" : ((s) < 1048576)? "k" : \
	((s) < 1073741824)? "M" : ((s) < 1099511627776)? "G" : "T")

void CCLDeviceParams::Dump(FILE *p_fw)
{
	fprintf(p_fw, "device(\'%s\', " PRIsizeB
		"B RAM, %.2f MHz, multiproc: %d, max-threads-block: %d, max-block-size: %dx%dx%d)\n",
		p_s_Name(),
		PRIsizeBparams(n_Memory_Size()),
		t_Properties().clockRate / 1e3f,
		n_Multiprocessor_Num(),
		t_Properties().maxThreadsPerBlock,
		t_Properties().maxThreadsDim[0],
		t_Properties().maxThreadsDim[1],
		t_Properties().maxThreadsDim[2]);
	// show some device parameters
}

CCLDeviceParams::CCLDeviceParams(cl_device_id h_device)
	:m_h_device(h_device), m_n_memory_size(0)
{
	memset(m_p_device_caps, 0, 2 * sizeof(int));
	memset(&m_t_devprop, 0, sizeof(CLdevprop));

	if(!QueryDeviceParams()) {
		m_h_device = 0; // mark error
		return;
	}
	// get device params
}

CCLDeviceParams::CCLDeviceParams(cl_context h_context, int n_device_index)
	:m_h_device(0), // mark error
	m_n_memory_size(0)
{
	memset(m_p_device_caps, 0, 2 * sizeof(int));
	memset(&m_t_devprop, 0, sizeof(CLdevprop));

	{
		std::vector<cl_device_id> device_list;
		if(CCLUtils::n_GetDeviceList(h_context, device_list) != CL_SUCCESS)
			return;
		// get all the devices

		if(n_device_index < 0 || unsigned(n_device_index) >= device_list.size())
			return;
		// check index

		m_h_device = device_list[n_device_index];
	}
	// query all devices available to the context, get device handle

	if(!QueryDeviceParams()) {
		m_h_device = 0; // mark error
		return;
	}
	// get device params
}

bool CCLDeviceParams::QueryDeviceParams()
{
	if(n_GetDeviceInfoString(m_s_name, CL_DEVICE_NAME) != CL_SUCCESS)
		return false;
	// get device name

	bool b_nv_device_attribute_query = b_IsExtensionSupported("cl_nv_device_attribute_query");
	// determines wheter we have cl_nv_device_attribute_query

	{
		/*std::string s_version;
		if(n_GetDeviceInfoString(s_version, CL_DEVICE_VERSION) != CL_SUCCESS)
			return;
		m_p_device_caps[0] = atol(s_version.c_str());
		m_p_device_caps[1] = (strchr(s_version.c_str(), '.'))? atol(strchr(s_version.c_str(), '.') + 1) : 0;*/ // todo

		if(b_nv_device_attribute_query) {
			cl_uint n_major, n_minor;
			if(clGetDeviceInfo(m_h_device, CL_DEVICE_COMPUTE_CAPABILITY_MAJOR_NV, sizeof(cl_uint), &n_major, NULL) != CL_SUCCESS ||
			   clGetDeviceInfo(m_h_device, CL_DEVICE_COMPUTE_CAPABILITY_MINOR_NV, sizeof(cl_uint), &n_minor, NULL) != CL_SUCCESS)
				return false;
			m_p_device_caps[0] = n_major;
			m_p_device_caps[1] = n_minor;
		} else {
			m_p_device_caps[0] = 1;
			m_p_device_caps[1] = 0;
		}
	}
	// get device OpenCL revision

	{
		if(b_nv_device_attribute_query) {
			cl_uint n_value;
			if(clGetDeviceInfo(m_h_device, CL_DEVICE_KERNEL_EXEC_TIMEOUT_NV, sizeof(n_value), &n_value, NULL) != CL_SUCCESS)
				return false;
			m_b_kernel_exec_timeout = (n_value != 0);
		} else
			m_b_kernel_exec_timeout = false;
	}
	{
		cl_uint n_value;
		if(clGetDeviceInfo(m_h_device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(n_value), &n_value, NULL) != CL_SUCCESS)
			return false;
		m_n_multiprocessor_num = n_value;
	}
	// get some interesting device attributes

	{
		cl_ulong n_value;
		if(clGetDeviceInfo(m_h_device, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(n_value), &n_value, NULL) != CL_SUCCESS)
			return false;
		_ASSERTE(n_value <= UINT_MAX); // this might get crossed pretty soon
		m_n_memory_size = unsigned int(n_value);
	}
	// get device memory size

	return n_Query_DeviceProperties(m_t_devprop, m_h_device) == CL_SUCCESS;
}

int CCLDeviceParams::n_Query_DeviceProperties(CLdevprop &r_t_devprop, cl_device_id h_device)
{
	memset(&r_t_devprop, 0, sizeof(CLdevprop));
	// clear output

	bool b_nv_device_attribute_query = b_IsExtensionSupported(h_device, "cl_nv_device_attribute_query");
	// determines wheter we have cl_nv_device_attribute_query

	{
		cl_uint n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		r_t_devprop.clockRate = n_value * 1000;
	}
	{
		cl_uint n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_ADDRESS_BITS, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		n_value /= 2;
		int n_size = int(std::min(uint64_t(INT_MAX), (uint64_t(1) << (n_value - 1)) | ((uint64_t(1) << (n_value - 1)) - 1)));
		r_t_devprop.maxGridSize[0] = n_size;
		r_t_devprop.maxGridSize[1] = n_size;
		r_t_devprop.maxGridSize[2] = 1; // this is wrong; todo
	}
	{
		size_t p_value[3]; // 3 is safe value, no device shall have CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS less than 3
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(p_value), p_value, NULL)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(p_value[0] <= INT_MAX && p_value[1] <= INT_MAX && p_value[2] <= INT_MAX);
		r_t_devprop.maxThreadsDim[0] = int(p_value[0]);
		r_t_devprop.maxThreadsDim[1] = int(p_value[1]);
		r_t_devprop.maxThreadsDim[2] = int(p_value[2]); // block size
	}
	{
		size_t n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(n_value <= INT_MAX);
		r_t_devprop.maxThreadsPerBlock = int(n_value);
	}
	{
		cl_ulong n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(n_value <= INT_MAX);
		r_t_devprop.memPitch = int(n_value);
	}
	{
		if(b_nv_device_attribute_query) {
			cl_uint n_value;
			int n_result;
			if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_REGISTERS_PER_BLOCK_NV, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
				return n_result;
			r_t_devprop.regsPerBlock = n_value;
		} else
			r_t_devprop.regsPerBlock = -1; // unknown
	}
	{
		cl_ulong n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_LOCAL_MEM_SIZE, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(n_value <= INT_MAX);
		r_t_devprop.sharedMemPerBlock = int(n_value);
	}
	{
		if(b_nv_device_attribute_query) {
			cl_uint n_value;
			int n_result;
			if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_WARP_SIZE_NV, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
				return n_result;
			r_t_devprop.SIMDWidth = n_value;
		} else
			r_t_devprop.SIMDWidth = -1; // unknown
	}
	{
		r_t_devprop.textureAlign = -1; // unknown; todo
	}
	{
		cl_ulong n_value;
		int n_result;
		if((n_result = clGetDeviceInfo(h_device, CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE, sizeof(n_value), &n_value, NULL)) != CL_SUCCESS)
			return n_result;
		_ASSERTE(n_value <= INT_MAX);
		r_t_devprop.totalConstantMemory = int(n_value);
	}
	// get device properties

	return CL_SUCCESS;
}

bool CCLDeviceParams::b_ProblemFitsAtOnce(int n_width, int n_height, int n_depth) const
{
	n_width = (n_width + m_t_devprop.maxGridSize[0] - 1) / m_t_devprop.maxGridSize[0];
	n_height = (n_height + m_t_devprop.maxGridSize[1] - 1) / m_t_devprop.maxGridSize[1];
	n_depth = (n_depth + m_t_devprop.maxGridSize[2] - 1) / m_t_devprop.maxGridSize[2];
	// calculate dimensions, relative to maximal grid size (round up)

	if(n_width > m_t_devprop.maxThreadsDim[0] ||
	   n_height > m_t_devprop.maxThreadsDim[1] ||
	   n_depth > m_t_devprop.maxThreadsDim[2])
		return false;
	// those dimensions must not exceed block size

	if(n_width * n_height * n_depth > m_t_devprop.maxThreadsPerBlock)
		return false;
	// number of threads in the block must be below limit

	return true;
}

bool CCLDeviceParams::CalculateGridParams(int *p_block_size, int *p_grid_size,
	int n_width, int n_height, int n_depth) const
{
	if(!b_ProblemFitsAtOnce(n_width, n_height, n_depth))
		return false;
	// @todo - handle subdivided problems too

	int n_blk_width = (n_width + m_t_devprop.maxGridSize[0] - 1) / m_t_devprop.maxGridSize[0];
	int n_blk_height = (n_height + m_t_devprop.maxGridSize[1] - 1) / m_t_devprop.maxGridSize[1];
	int n_blk_depth = (n_depth + m_t_devprop.maxGridSize[2] - 1) / m_t_devprop.maxGridSize[2];
	// calculate block dimensions (lower bound)

	// @todo - optimize block dimensions to approach m_t_devprop.maxThreadsPerBlock as closely as possible (that is good thing to do, right?)

	p_block_size[0] = n_blk_width;
	p_block_size[1] = n_blk_height;
	p_block_size[2] = n_blk_depth;
	p_grid_size[0] = (n_width + n_blk_width - 1) / n_blk_width;
	p_grid_size[1] = (n_height + n_blk_height - 1) / n_blk_height;
	p_grid_size[2] = (n_depth + n_blk_depth - 1) / n_blk_depth;
	// store block sizes, and grid size

	return true;
}

/*
 *								=== ~CCLDeviceParams ===
 */

/*
 *								=== CCLProgramStorage::_TSHA1 ===
 */

inline bool CCLProgramStorage::_TSHA1::operator ==(const _TSHA1 &r_t_other) const
{
	return !memcmp(p_data, r_t_other.p_data, 5 * sizeof(uint32_t));
}

/**
 *	@brief inequality comparator
 *	@param[in] r_t_other is hash to compare against
 *	@return Returns false if this hash equals r_t_other, otherwise returns true.
 */
inline bool CCLProgramStorage::_TSHA1::operator !=(const _TSHA1 &r_t_other) const
{
	return !(*this == r_t_other);
}

/*
 *								=== ~CCLProgramStorage::_TSHA1 ===
 */

/*
 *								=== CCLProgramStorage::TProgramInstance ===
 */

bool CCLProgramStorage::TProgramInstance::Read(FILE *p_fr, size_t n_device_num)
{
	_ASSERTE(!t_data_buffer.p_data && binary_size_list.empty()); // make sure this is empty

	if(fread(&t_build_opts_hash, sizeof(_TSHA1), 1, p_fr) != 1)
		return false;
	// read build options hash

	std::vector<uint64_t> binary_size_list64;
	try {
		binary_size_list64.resize(n_device_num);
		binary_size_list.resize(n_device_num);
	} catch(std::bad_alloc) {
		return false;
	}
	// alloc arrays

	if(fread(&binary_size_list64[0], sizeof(uint64_t), n_device_num, p_fr) != n_device_num)
		return false;
	uint64_t n_size_total = 0;
	for(size_t i = 0; i < n_device_num; ++ i) {
		if(binary_size_list64[i] > SIZE_MAX || n_size_total > UINT64_MAX - binary_size_list64[i])
			return false;
		binary_size_list[i] = size_t(binary_size_list64[i]);
		n_size_total += binary_size_list64[i];
	}
	// read binary lengths, sum up

	t_data_buffer.n_size = n_size_total;
	if(n_size_total > SIZE_MAX || !(t_data_buffer.p_data = new(std::nothrow) uint8_t[size_t(n_size_total)]))
		return false;
	{
		uint64_t n_size_remaining = n_size_total;
		uint8_t *p_data_ptr = t_data_buffer.p_data;
		while(n_size_remaining > 0) {
			int n_read = int(std::min(uint64_t(INT_MAX), n_size_remaining));
			if(fread(p_data_ptr, 1, n_read, p_fr) != n_read)
				return false;
			n_size_remaining -= n_read;
			p_data_ptr += n_read;
		}
		_ASSERTE(p_data_ptr == t_data_buffer.p_data + n_size_total);
	}
	// alloc and read data buffer

	return true;
}

bool CCLProgramStorage::TProgramInstance::Download(cl_program h_program, size_t n_device_num, _TSHA1 t_build_options_hash)
{
	_ASSERTE(!t_data_buffer.p_data && binary_size_list.empty()); // make sure this is empty

	t_build_opts_hash = t_build_options_hash;
	// copy hash of build options

	cl_uint _n_device_num;
	if(clGetProgramInfo(h_program, CL_PROGRAM_NUM_DEVICES,
	   sizeof(size_t), &_n_device_num, NULL) != CL_SUCCESS || _n_device_num != n_device_num)
		return false;
	// get number of devices, make sure it matches description

	if(!n_device_num || n_device_num > UINT32_MAX)
		return false;
	// we can't query binaries without devices

	std::vector<const unsigned char*> binary_ptr_list;
	try {
		binary_ptr_list.resize(n_device_num);
#ifdef _DEBUG
		_ASSERTE(n_device_num < SIZE_MAX);
		binary_size_list.resize(n_device_num + 1);
		if(n_device_num)
			binary_size_list[n_device_num - 1] = 0xbaadf00d; // paranoid check
		binary_size_list[n_device_num] = 0xbaadf00d; // "magic" word
#else //_DEBUG
		binary_size_list.resize(n_device_num);
#endif //_DEBUG
	} catch(std::bad_alloc&) {
		return false;
	}
	if(clGetProgramInfo(h_program, CL_PROGRAM_BINARY_SIZES,
	   n_device_num * sizeof(size_t), &binary_size_list[0], NULL) != CL_SUCCESS)
		return false;
#ifdef _DEBUG
	_ASSERTE(binary_size_list[n_device_num] == 0xbaadf00d); // make sure it's returning the right amount of data
	_ASSERTE(!n_device_num || binary_size_list[n_device_num - 1] != 0xbaadf00d); // paranoid check
	binary_size_list.erase(binary_size_list.end() - 1); // remove it from the list, length must be equal to n_device_num
#endif //_DEBUG
	// get binary sizes

	uint64_t n_size_total = 0;
	for(size_t i = 0; i < n_device_num; ++ i) {
		if(n_size_total > UINT64_MAX - binary_size_list[i])
			return false;
		n_size_total += binary_size_list[i];
	}
	// sum binary sizes up

	t_data_buffer.n_size = n_size_total;
	if(n_size_total > SIZE_MAX || !(t_data_buffer.p_data = new(std::nothrow) uint8_t[size_t(n_size_total)]))
		return false;
	// alloc data buffer

	{
		const uint8_t *p_data_ptr = t_data_buffer.p_data;
		for(size_t i = 0; i < n_device_num; ++ i) {
			binary_ptr_list[i] = (const unsigned char*)p_data_ptr;
			p_data_ptr += binary_size_list[i];
		}
		_ASSERTE(p_data_ptr == t_data_buffer.p_data + t_data_buffer.n_size);
	}
	// get pointers to individual binaries

	if(clGetProgramInfo(h_program, CL_PROGRAM_BINARIES, size_t(n_size_total), &binary_ptr_list[0], NULL) != CL_SUCCESS)
		return false;
	// download binaries

	return true;
}

bool CCLProgramStorage::TProgramInstance::Write(FILE *p_fw, size_t n_device_num)
{
	if(binary_size_list.size() != n_device_num)
		return false;
	// make sure there is correct number of binaries

	if(fwrite(&t_build_opts_hash, sizeof(_TSHA1), 1, p_fw) != 1)
		return false;
	// write build options hash

	for(size_t i = 0; i < n_device_num; ++ i) {
		_ASSERTE(SIZE_MAX <= UINT64_MAX);
		uint64_t n_size = binary_size_list[i];
		if(fwrite(&n_size, sizeof(uint64_t), 1, p_fw) != 1)
			return false;
	}
	// write binary lengths

	{
		uint64_t n_size_remaining = t_data_buffer.n_size;
		uint8_t *p_data_ptr = t_data_buffer.p_data;
		while(n_size_remaining) {
			int n_write = int(std::min(uint64_t(INT_MAX), n_size_remaining));
			if(fwrite(p_data_ptr, 1, n_write, p_fw) != n_write)
				return false;
			n_size_remaining -= n_write;
			p_data_ptr += n_write;
		}
		_ASSERTE(p_data_ptr == t_data_buffer.p_data + t_data_buffer.n_size);
	}
	// write binaries data buffer

	return true;
}

void CCLProgramStorage::TProgramInstance::Free()
{
	if(t_data_buffer.p_data) {
		delete[] t_data_buffer.p_data;
		t_data_buffer.p_data = 0;
	}
	binary_size_list.clear();
}

inline void CCLProgramStorage::TProgramInstance::Swap(TProgramInstance &r_t_other)
{
	std::swap(t_build_opts_hash, r_t_other.t_build_opts_hash);
	binary_size_list.swap(r_t_other.binary_size_list);
	std::swap(t_data_buffer, r_t_other.t_data_buffer);
}

/*
 *								=== ~CCLProgramStorage::TProgramInstance ===
 */

/*
 *								=== CCLProgramStorage ===
 */

CCLProgramStorage::CCLProgramStorage()
	:m_b_dirty(false)
{
	memset(&m_t_header, 0, sizeof(TFileHeader));

	m_t_header.p_magic[0] = 'L';
	m_t_header.p_magic[1] = 'C';
	m_t_header.p_magic[2] = 'L';
	m_t_header.p_magic[3] = 's';
	// set magic word ("Lame CL storage") - so b_Status() doesn't fail
}

CCLProgramStorage::CCLProgramStorage(const char *p_s_filename)
	:m_b_dirty(false)
{
	if(!Load(p_s_filename))
		memset(&m_t_header, 0, sizeof(TFileHeader)); // to mark error
}

CCLProgramStorage::~CCLProgramStorage()
{
	std::for_each(instance_list.begin(), instance_list.end(), DeleteInstance);
}

bool CCLProgramStorage::b_Status() const
{
	return m_t_header.p_magic[0] == 'L' && m_t_header.p_magic[1] == 'C' &&
		m_t_header.p_magic[2] == 'L' && m_t_header.p_magic[3] != 's';
}

bool CCLProgramStorage::b_Dirty() const
{
	/*_TSHA1 t_ctrl_hash;
	if(!Calc_ControlHash(t_ctrl_hash))
		fprintf(stderr, "hash check failed ...\n"); // debug
	if(m_b_dirty != (t_ctrl_hash != m_t_header.t_control_hash)) {
		fprintf(stderr, "dirty discrepancy (dirty %d, hash-check %d) ...\n",
			m_b_dirty, t_ctrl_hash != m_t_header.t_control_hash); // debug
	}*/ // debug

	return m_b_dirty;
}

bool CCLProgramStorage::b_Check_DeviceIds(size_t n_device_num, const cl_device_id *p_device_list) const
{
	if(device_id_list.size() != n_device_num)
		return false;
	// counts must match

	std::string s_tmp0, s_tmp1;
	for(size_t i = 0, n = device_id_list.size(); i < n; ++ i) {
		if(CCLDeviceParams::n_GetDeviceInfoString(s_tmp0,
		   p_device_list[i], CL_DEVICE_NAME) != CL_SUCCESS ||
		   CCLDeviceParams::n_GetDeviceInfoString(s_tmp1,
		   p_device_list[i], CL_DRIVER_VERSION) != CL_SUCCESS )
			return false; // this is, in fact, error, not npos||
		try {
			s_tmp0 += "|";
			s_tmp0 += s_tmp1;
		} catch(std::bad_alloc&) {
			return false;
		}
		if(s_tmp0.compare(device_id_list[i]))
			return false;
	}
	// check device identification

	return true;
}

size_t CCLProgramStorage::n_Find_ProgramBinaries(const char *p_s_source_code, size_t n_device_num,
	const cl_device_id *p_device_list, const char *p_s_build_options) const
{
	if(device_id_list.size() != n_device_num ||
	   m_t_header.t_source_hash != t_Hash_String(p_s_source_code) ||
	   !b_Check_DeviceIds(n_device_num, p_device_list))
		return npos;
	// make sure there are the same devices (that's kind of tricky though, this is limited
	// to number of devices as there's no device id that could be written to the file (except
	// maybe hash of name and drivers)) and the same source code // @t_odo - implement better device identification

	_TSHA1 t_build_opts_hash = t_Hash_String(p_s_build_options);
	// hash build options

	for(size_t i = 0, n = instance_list.size(); i < n; ++ i) {
		if(instance_list[i].t_build_opts_hash == t_build_opts_hash) {
			if(i) {
				/*fprintf(stderr, "promoting binaries %d\n", i);*/ // debug
				instance_list[i].Swap(instance_list[i - 1]);
				-- i;
				// promote this instance, effectively invalidating any indices obtained by previous calls to this functions

				m_b_dirty = true;
				// we just changed order of binaries
			}
			/*fprintf(stderr, "requested binaries found: %d (\'%s\')\n", i, p_s_build_options);*/ // debug
			return i;
		}
	}
	// try to find the program binary

	/*fprintf(stderr, "requested binaries not found (\'%s\')\n", p_s_build_options);*/ // debug

	return npos;
	// not found
}

int CCLProgramStorage::n_Get_ProgramBinaries(cl_context h_context, cl_program *p_program, cl_int &r_n_result,
	size_t n_binaries_index, size_t n_device_num, const cl_device_id *p_device_list) const
{
	if(n_binaries_index == npos || n_binaries_index >= instance_list.size() || device_id_list.size() != n_device_num)
		return build_InvalidParams;
	// check binaries index and devices

	_ASSERTE(b_Check_DeviceIds(n_device_num, p_device_list));
	// device identification isn't checked in release, it's checked in n_Find_ProgramBinaries(),
	// this function is supposed to get the same parameters

	const TProgramInstance &r_t_inst = instance_list[n_binaries_index];
	// get program instance

	std::vector<const unsigned char*> binary_ptr_list;
	try {
		binary_ptr_list.resize(n_device_num);
	} catch(std::bad_alloc&) {
		return build_LowMemory;
	}
	{
		const uint8_t *p_data_ptr = r_t_inst.t_data_buffer.p_data;
		for(size_t i = 0; i < n_device_num; ++ i) {
			binary_ptr_list[i] = (const unsigned char*)p_data_ptr;
			p_data_ptr += r_t_inst.binary_size_list[i];
		}
		_ASSERTE(p_data_ptr == r_t_inst.t_data_buffer.p_data + r_t_inst.t_data_buffer.n_size);
	}
	// get pointers to individual binaries

	cl_int n_result;
	_ASSERTE(n_device_num <= /*CL*/UINT_MAX);
	cl_program h_prog = clCreateProgramWithBinary(h_context, cl_uint(n_device_num), p_device_list,
		&r_t_inst.binary_size_list[0], &binary_ptr_list[0], NULL, &n_result);
	if(n_result != CL_SUCCESS) {
		r_n_result = n_result;
		return build_CreateProgramFailure;
	}
	if((n_result = clBuildProgram(h_prog, 0, NULL, NULL, NULL, NULL)) != CL_SUCCESS) { // note no build options here
		r_n_result = n_result;
		return build_BuildProgramFailure;
	}
	// use loaded binaries to create the program ...

	r_n_result = CL_SUCCESS; // !!
	*p_program = h_prog;
	// write output

	return 0;
	// success
}

bool CCLProgramStorage::Put_ProgramBinaries(cl_program h_program, const char *p_s_source_code,
	size_t n_device_num, const cl_device_id *p_device_list, const char *p_s_build_options)
{
	// puts compiled program binaries to the storage
	// number of instances is limited (LRU-type policy), but not here
	// note this might erase all other instances in case source code changed

	for(size_t i = 0; i < n_device_num; ++ i) {
		cl_build_status n_status;
		if(clGetProgramBuildInfo(h_program, p_device_list[i], CL_PROGRAM_BUILD_STATUS,
		   sizeof(cl_build_status), &n_status, NULL) != CL_SUCCESS || n_status != CL_BUILD_SUCCESS)
			return false;
	}
	// the program must be completely and successfuly built

	if(n_Find_ProgramBinaries(p_s_source_code, n_device_num, p_device_list, p_s_build_options) != npos)
		return true;
	// it's already there

	m_b_dirty = true;
	// we're going to modify the binaries

	_TSHA1 t_source_hash = t_Hash_String(p_s_source_code);
	// calculate hash of source code

	if(device_id_list.size() != n_device_num || m_t_header.t_source_hash != t_source_hash ||
	   !b_Check_DeviceIds(n_device_num, p_device_list)) { // this is costy, make it at the end
		std::for_each(instance_list.begin(), instance_list.end(), DeleteInstance);
		instance_list.clear();
		device_id_list.clear();

		m_t_header.t_source_hash = t_source_hash;

		try {
			device_id_list.resize(n_device_num);
		} catch(std::bad_alloc&) {
			return false;
		}
		for(size_t i = 0; i < n_device_num; ++ i) {
			std::string s_dev_driver;
			if(CCLDeviceParams::n_GetDeviceInfoString(device_id_list[i],
			   p_device_list[i], CL_DEVICE_NAME) != CL_SUCCESS ||
			   CCLDeviceParams::n_GetDeviceInfoString(s_dev_driver,
			   p_device_list[i], CL_DRIVER_VERSION) != CL_SUCCESS) {
				device_id_list.clear();
				return false;
			}
			try {
				device_id_list[i] += "|";
				device_id_list[i] += s_dev_driver;
			} catch(std::bad_alloc&) {
				device_id_list.clear();
				return false;
			}
		}
		// get list of device id's (supposedly unique enough id's)
	}
	// in case it is completely different program, it invalidates all the instances

	try {
		instance_list.insert(instance_list.begin(), 1, TProgramInstance()); // add new instances at the beginning of the list
	} catch(std::bad_alloc&) {
		return false;
	}
	_TSHA1 t_build_opts_hash = t_Hash_String(p_s_build_options);
	if(!instance_list.front().Download(h_program, n_device_num, t_build_opts_hash)) {
		instance_list.front().Free();
		instance_list.erase(instance_list.begin()); // try to clean up so this would still be useful
		return false;
	}
	// add and fill a new program instance

	return true;
}

bool CCLProgramStorage::Save(const char *p_s_filename, size_t n_max_program_instance_num)
{
	if(!Calc_ControlHash(m_t_header.t_control_hash, n_max_program_instance_num))
		return false;
	// update hash

	FILE *p_fw;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fw, p_s_filename, "wb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fw = fopen(p_s_filename, "wb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		return false;
	// open file

	m_t_header.p_magic[0] = 'L';
	m_t_header.p_magic[1] = 'C';
	m_t_header.p_magic[2] = 'L';
	m_t_header.p_magic[3] = 's';
	// set magic word ("Lame CL storage")

	_ASSERTE(device_id_list.size() < UINT32_MAX);
	uint32_t n_device_num = uint32_t(device_id_list.size());
	// set device count

	if(std::min(n_max_program_instance_num, instance_list.size()) > UINT32_MAX) {
		fclose(p_fw);
		return false;
	}
	uint32_t n_instance_num = uint32_t(std::min(n_max_program_instance_num, instance_list.size()));
	if(fwrite(&m_t_header, sizeof(TFileHeader), 1, p_fw) != 1 ||
	   fwrite(&n_device_num, sizeof(uint32_t), 1, p_fw) != 1 ||
	   fwrite(&n_instance_num, sizeof(uint32_t), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write header

	for(size_t i = 0, n = device_id_list.size(); i < n; ++ i) {
		_ASSERTE(device_id_list[i].length() < SIZE_MAX);
		_ASSERTE(sizeof(char) == 1); // this better be consistent
		size_t n_length = device_id_list[i].length();
		if(fwrite(device_id_list[i].c_str(), sizeof(char), n_length + 1, p_fw) != n_length + 1) {
			fclose(p_fw);
			return false;
		}
	}
	// write device names (null-terminated)

	for(size_t i = 0; i < n_instance_num; ++ i) {
		if(!instance_list[i].Write(p_fw, device_id_list.size())) {
			fclose(p_fw);
			return false;
		}
	}
	// write instances

	if(ferror(p_fw)) {
		fclose(p_fw);
		return false;
	}
	fclose(p_fw);
	// close file

	m_b_dirty = (n_max_program_instance_num < instance_list.size());
	// the file now contains actual data (in case we didn't trim it)

	/*fprintf(stderr, "saving binaries ...\n");*/ // debug

	return true;
}

bool CCLProgramStorage::Load(const char *p_s_filename)
{
	/*fprintf(stderr, "loading binaries ...\n");*/ // debug

	std::for_each(instance_list.begin(), instance_list.end(), DeleteInstance);
	instance_list.clear();
	device_id_list.clear();
	// clear first

	FILE *p_fr;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fr, p_s_filename, "rb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fr = fopen(p_s_filename, "rb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		return false;
	// open file

	uint32_t n_instance_num, n_device_num;
	if(fread(&m_t_header, sizeof(TFileHeader), 1, p_fr) != 1 ||
	   fread(&n_device_num, sizeof(uint32_t), 1, p_fr) != 1 ||
	   fread(&n_instance_num, sizeof(uint32_t), 1, p_fr) != 1 ||
	   n_device_num > SIZE_MAX || n_instance_num > SIZE_MAX ||
	   m_t_header.p_magic[0] != 'L' || m_t_header.p_magic[1] != 'C' ||
	   m_t_header.p_magic[2] != 'L' || m_t_header.p_magic[3] != 's') {
		fclose(p_fr);
		return false;
	}
	// read header

	try {
		_ASSERTE(n_device_num <= SIZE_MAX);
		device_id_list.resize(size_t(n_device_num));
		_ASSERTE(n_instance_num <= SIZE_MAX);
		instance_list.resize(size_t(n_instance_num));
	} catch(std::bad_alloc&) {
		fclose(p_fr);
		return false;
	}
	// alloc devices and instances

	{
		for(size_t i = 0, n = device_id_list.size(); i < n; ++ i) {
			std::string s_id;
			for(;;) {
				char c;
				if(fread(&c, sizeof(char), 1, p_fr) != 1) {
					device_id_list.clear();
					return false;
				}
				// read a single character

				if(!c)
					break;
				// detect terminating null

				try {
					s_id.push_back(c);
				} catch(std::bad_alloc&) {
					device_id_list.clear();
					fclose(p_fr);
					return false;
				}
				// add character to the string
			}

			device_id_list[i].swap(s_id);
		}
	}
	// read device id's

	for(size_t i = 0, n = instance_list.size(); i < n; ++ i) {
		if(!instance_list[i].Read(p_fr, size_t(n_device_num))) {
			fclose(p_fr);

			std::for_each(instance_list.begin(), instance_list.end(), DeleteInstance);
			instance_list.clear();
			device_id_list.clear();
			// !!

			return false;
		}
	}
	// read instances

	fclose(p_fr);
	// close file

	_TSHA1 t_ctrl_hash;
	if(!Calc_ControlHash(t_ctrl_hash) || t_ctrl_hash != m_t_header.t_control_hash) {
		/*fprintf(stderr, "hash check failed ...\n");*/ // debug

		std::for_each(instance_list.begin(), instance_list.end(), DeleteInstance);
		instance_list.clear();
		device_id_list.clear();
		// !!

		return false;
	}
	// check hash

	m_b_dirty = false;
	// actual data are now loaded

	return true;
}

CCLProgramStorage::_TSHA1 CCLProgramStorage::t_Hash_String(const char *p_s_str)
{
	CStreamHash<TSHA1> hash;
	hash.Process_Data(p_s_str, strlen(p_s_str) * sizeof(char));
	// calculate

	TSHA1 t_hash = hash.t_Result();
	_TSHA1 t_result;
	memcpy(&t_result, &t_hash[0], sizeof(_TSHA1)); // TSHA1 contains byte counter which is not needed
	// convert result

	return t_result;
}

bool CCLProgramStorage::Calc_ControlHash(_TSHA1 &r_t_result, size_t n_instance_limit) const
{
	CStreamHash<TSHA1> hash;
	hash.Process_Data(&m_t_header.t_source_hash, sizeof(_TSHA1));
	_ASSERTE(device_id_list.size() <= UINT32_MAX);
	uint32_t n_device_num = uint32_t(device_id_list.size());
	hash.Process_Data(&n_device_num, sizeof(uint32_t));
	_ASSERTE(instance_list.size() <= UINT32_MAX);
	uint32_t n_instance_num = uint32_t(instance_list.size());
	hash.Process_Data(&n_instance_num, sizeof(uint32_t));
	// hash header ...

	for(size_t i = 0, n = device_id_list.size(); i < n; ++ i) {
		_ASSERTE(device_id_list[i].length() < SIZE_MAX);
		hash.Process_Data(device_id_list[i].c_str(), (device_id_list[i].length() + 1) * sizeof(char));
	}
	// hash device id's (including the terminating null - to mimic layout in the file)

	for(size_t i = 0, n = std::min(n_instance_limit, instance_list.size()); i < n; ++ i) {
		const TProgramInstance &r_t_inst = instance_list[i];
		// get program instance ...

		hash.Process_Data(&r_t_inst.t_build_opts_hash, sizeof(_TSHA1));
		if(r_t_inst.binary_size_list.size() != device_id_list.size())
			return false;
		for(size_t j = 0, m = r_t_inst.binary_size_list.size(); j < m; ++ j) {
			_ASSERTE(SIZE_MAX <= UINT64_MAX);
			uint64_t n_size = r_t_inst.binary_size_list[j];
			hash.Process_Data(&n_size, sizeof(uint64_t));
		}
		// hash header

		_ASSERTE(r_t_inst.t_data_buffer.n_size <= SIZE_MAX);
		hash.Process_Data(r_t_inst.t_data_buffer.p_data, size_t(r_t_inst.t_data_buffer.n_size)); // note this might overflow
		// hash data
	}
	// hash instances

	TSHA1 t_hash = hash.t_Result();
	memcpy(&r_t_result, &t_hash[0], sizeof(_TSHA1)); // TSHA1 contains byte counter which is not needed
	// write result

	return true;
}

inline void CCLProgramStorage::DeleteInstance(TProgramInstance &r_t_inst)
{
	r_t_inst.Free();
}

/*
 *								=== ~CCLProgramStorage ===
 */

/*
 *								=== CCLProgramCompiler ===
 */

#define CL_COMPILER_UTIL_BOGUS_DATA_DETECTION

void CCLProgramCompiler::Dump_StatusWord(int n_status_word, FILE *p_fw)
{
	if(n_status_word & cache_ReadAttempted) {
		if(n_status_word & cache_ReadSucceeded)
			fprintf(p_fw, "program loaded from cache\n");
		else {
			if(n_status_word & cache_ReadFailed_FileNotFound)
				fprintf(p_fw, "program not cached\n");
			else if(n_status_word & cache_ReadFailed_IO)
				fprintf(p_fw, "i/o error while loading program from cache\n");
			else if(n_status_word & cache_ReadFailed_OutOfMemory)
				fprintf(p_fw, "not enough memory while loading program from cache\n");
			else if(n_status_word & cache_ReadFailed_SourceChecksum)
				fprintf(p_fw, "program cache miss\n");
			else if(n_status_word & cache_ReadFailed_BinaryChecksum)
				fprintf(p_fw, "program cache corrupt\n");
			else if(n_status_word & cache_ReadFailed_CreateProgram)
				fprintf(p_fw, "program cached for different device\n");
			else
				fprintf(p_fw, "unknown error while loading program from cache\n");
		}
	}
	if(n_status_word & prog_CompiledFromSource) {
		if(!(n_status_word & prog_CreateSucceeded))
			fprintf(p_fw, "clCreateProgramWithSource() failed while compiling program from source\n");
		else if(!(n_status_word & prog_BuildSucceeded))
			fprintf(p_fw, "clBuildProgram() failed while compiling program from source\n");
		else
			fprintf(p_fw, "program was successfully compiled from source\n");
	}
	if(n_status_word & cache_WriteAttempted) {
		if(n_status_word & cache_WriteSucceeded) {
			if(n_status_word & cache_WriteTrimmed) {
				fprintf(p_fw, "program binaries were successfully written to cache, binary versions dropped\n");
			} else
				fprintf(p_fw, "program binaries were successfully written to cache\n");
		} else
			fprintf(p_fw, "failed to cache program binaries\n");
	}
}

int CCLProgramCompiler::n_CompileProgram(cl_context h_context, cl_program *p_program,
	const char *p_s_source, size_t n_device_num, const cl_device_id *p_device_list,
	const char *p_s_build_options, const char *p_s_cache_file, int *p_status_word,
	size_t n_max_program_instance_num)
{
	int n_dummy_status_word;
	if(!p_status_word)
		p_status_word = &n_dummy_status_word;
	// makes sure p_status_word is not null

	*p_status_word = 0;

#ifdef CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
	TSHA1 t_hash = t_Hash_ProgramSource_BuildOptions(p_s_source, p_s_build_options);
	// calculate program and build options hash
#else //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
	CCLProgramStorage program_storage;
	// prepare program storage
#endif //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT

	if(p_s_cache_file) {
		*p_status_word |= cache_ReadAttempted;

#ifdef CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
		int n_result;
		if((n_result = n_ReadProgramBinaries(p_s_cache_file, t_hash, h_context,
		   p_program, n_device_num, p_device_list)) == cache_ReadSucceeded) {
			*p_status_word |= cache_ReadSucceeded;
			return CL_SUCCESS;
		}
		// use the old function

		*p_status_word |= n_result;
#else //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
		if(program_storage.Load(p_s_cache_file)) {
			size_t n_binaries_index;
			if((n_binaries_index = program_storage.n_Find_ProgramBinaries(p_s_source,
			   n_device_num, p_device_list, p_s_build_options)) != CCLProgramStorage::npos) {
				int n_result; cl_int n_cl_result;
				if((n_result = program_storage.n_Get_ProgramBinaries(h_context, p_program, n_cl_result, n_binaries_index,
				   n_device_num, p_device_list)) == CCLProgramStorage::build_Success) {
					*p_status_word |= cache_ReadSucceeded;

					if(program_storage.b_Dirty()) {
						*p_status_word |= cache_WriteAttempted;

#ifdef CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
						std::string s_temp_name;
						char p_s_time[5];
						sprintf(p_s_time, "%04x", clock() & 0xffff); // "random" identifier
						try {
							s_temp_name = p_s_cache_file;
							s_temp_name += ".tmp";
							s_temp_name += p_s_time;
						} catch(std::bad_alloc&) {
							return CL_SUCCESS; // load succeeded, just the save failed. ignore it.
						}
						// create temp filename

						if(program_storage.Save(s_temp_name.c_str(), n_max_program_instance_num)) {
							remove(p_s_cache_file);
							rename(s_temp_name.c_str(), p_s_cache_file);

							*p_status_word |= cache_WriteSucceeded;
						}
						// save with new filename, if succeeds, rewrite the old file; this is important because of the lru policy
#else //CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
						if(program_storage.Save(p_s_cache_file, n_max_program_instance_num)) {
							*p_status_word |= cache_WriteSucceeded;
							if(program_storage.b_Dirty())
								*p_status_word |= cache_WriteTrimmed; // it's dirty - file doesn't contain all the versions - n_max_program_instance_num limit
						} else
							remove(p_s_cache_file); // don't leave garbage
						// save with old filename, remove file on failure (no fallback)
#endif //CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
					}

					return CL_SUCCESS;
				}
				switch(n_result) {
				case CCLProgramStorage::build_InvalidParams:
					*p_status_word |= cache_ReadFailed_OutOfMemory; // not actually the right error code (but it shouldn't happen in here, it's more of an internal program error)
					break;
				case CCLProgramStorage::build_LowMemory:
					*p_status_word |= cache_ReadFailed_OutOfMemory;
					break;
				case CCLProgramStorage::build_CreateProgramFailure:
					*p_status_word |= cache_ReadFailed_CreateProgram;
					break;
				case CCLProgramStorage::build_BuildProgramFailure:
					*p_status_word |= cache_ReadFailed_CreateProgram; // not actually the right error code
					break;
				}
			} else
				*p_status_word |= cache_ReadFailed_SourceChecksum;
		} else {
			if(access(p_s_cache_file, 00) != 0)
				*p_status_word |= cache_ReadFailed_FileNotFound;
			else
				*p_status_word |= cache_ReadFailed_IO;
		}
		// use the new program storage
#endif //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
	}
	// attempt to load program binaries from file

	cl_program h_prog;
	{
		*p_status_word |= prog_CompiledFromSource;

		cl_int n_result;
		h_prog = clCreateProgramWithSource(h_context, 1, &p_s_source, NULL, &n_result);
		if(n_result != CL_SUCCESS)
			return n_result;

		*p_status_word |= prog_CreateSucceeded;

		_ASSERTE(n_device_num <= /*CL*/UINT_MAX);
		if((n_result = clBuildProgram(h_prog, cl_uint(n_device_num), p_device_list, p_s_build_options, NULL, NULL)) != CL_SUCCESS) {
#ifdef CL_PROGRAM_COMPILER_DISPLAY_BUILD_ERRORS
			if(n_device_num) {
				std::string s_build_log, s_device_name;
				for(size_t i = 0; i < n_device_num; ++ i) {
					cl_build_status n_build_status;

					if(Get_BuildLog(s_build_log, n_build_status, h_prog, p_device_list[i]) != CL_SUCCESS)
						return n_result;
					// get buld log and build status

					if(CCLDeviceParams::n_GetDeviceInfoString(s_device_name, p_device_list[i], CL_DEVICE_NAME) != CL_SUCCESS)
						s_device_name.erase();
					const char *p_s_device_name = (s_device_name.empty())? "(null)" : s_device_name.c_str();
					// get device name

					if(n_build_status == CL_BUILD_NONE) {
						fprintf(stderr, "error: program wasn't built for device %d (%s)\n", i, p_s_device_name);
						continue;
					} else if(n_build_status == CL_BUILD_ERROR) {
						fprintf(stderr, "error: there were errors while building program for device %d (%s)\n",
							i, p_s_device_name);
					} else if(n_build_status != CL_BUILD_SUCCESS)
						fprintf(stderr, "error: unknown program build status for device %d (%s)\n", i, p_s_device_name);
					// show build status

					if(n_build_status != CL_BUILD_SUCCESS || !s_build_log.empty()) {
						fprintf(stderr, "=== OpenCL build log for device %d (%s) ===\n%s\n",
							i, p_s_device_name, s_build_log.c_str());
					}
					// show build log
				}
			}
			// show build error(s)
#endif //CL_PROGRAM_COMPILER_DISPLAY_BUILD_ERRORS

			return n_result;
		}
		// build program

#ifdef CL_PROGRAM_COMPILER_DISPLAY_SUCCESSFUL_BUILD_WARNINGS
		if(n_device_num) {
			std::string s_build_log, s_device_name;
			for(size_t i = 0; i < n_device_num; ++ i) {
				cl_build_status n_build_status;

				if(Get_BuildLog(s_build_log, n_build_status, h_prog, p_device_list[i]) != CL_SUCCESS)
					return n_result;
				// get buld log and build status

				if(CCLDeviceParams::n_GetDeviceInfoString(s_device_name, p_device_list[i], CL_DEVICE_NAME) != CL_SUCCESS)
					s_device_name.erase();
				const char *p_s_device_name = (s_device_name.empty())? "(null)" : s_device_name.c_str();
				// get device name

				if(n_build_status == CL_BUILD_NONE) {
					fprintf(stderr, "warning: program wasn't built for device %d (%s)\n", i, p_s_device_name);
					continue;
				} else if(n_build_status == CL_BUILD_ERROR) {
					fprintf(stderr, "warning: there were errors while building program for device %d (%s)\n",
						i, p_s_device_name);
				} else if(n_build_status != CL_BUILD_SUCCESS)
					fprintf(stderr, "warning: unknown program build status for device %d (%s)\n", i, p_s_device_name);
				// show build status

				if(n_build_status != CL_BUILD_SUCCESS || !s_build_log.empty()) {
					fprintf(stderr, "=== OpenCL build log for device %d (%s) ===\n%s\n",
						i, p_s_device_name, s_build_log.c_str());
				}
				// show build log
			}
		}
		// show warnings

		*p_status_word |= prog_BuildSucceeded;
	}
	// create program
#endif //CL_PROGRAM_COMPILER_DISPLAY_SUCCESSFUL_BUILD_WARNINGS

	*p_program = h_prog;
	// output

	if(p_s_cache_file) {
		*p_status_word |= cache_WriteAttempted;

#ifdef CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
		if(WriteProgramBinaries(h_prog, t_hash, p_s_cache_file))
			*p_status_word |= cache_WriteSucceeded;
		else {
			remove(p_s_cache_file);
			// in case it didn't save correctly, do not leave it arround
		}
#else //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
		if(program_storage.Put_ProgramBinaries(h_prog, p_s_source,
		   n_device_num, p_device_list, p_s_build_options)) {
			_ASSERTE(program_storage.b_Dirty());
			// should be "dirty" now

#ifdef CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
			std::string s_temp_name;
			char p_s_time[5];
			sprintf(p_s_time, "%04x", clock() & 0xffff); // "random" identifier
			try {
				s_temp_name = p_s_cache_file;
				s_temp_name += ".tmp";
				s_temp_name += p_s_time;
			} catch(std::bad_alloc&) {
				return CL_SUCCESS; // load succeeded, just the save failed. ignore it.
			}
			// create temp filename

			if(program_storage.Save(s_temp_name.c_str(), n_max_program_instance_num)) {
				remove(p_s_cache_file);
				rename(s_temp_name.c_str(), p_s_cache_file);

				*p_status_word |= cache_WriteSucceeded;
			}
			// save with new filename, if succeeds, rewrite the old file; this is important because of the lru policy
#else //CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
			if(program_storage.Save(p_s_cache_file, n_max_program_instance_num)) {
				*p_status_word |= cache_WriteSucceeded;
				if(program_storage.b_Dirty())
					*p_status_word |= cache_WriteTrimmed; // it's dirty - file doesn't contain all the versions - n_max_program_instance_num limit
			} else
				remove(p_s_cache_file); // don't leave garbage
			// save with old filename, remove file on failure (no fallback)
#endif //CL_PROGRAM_COMPILER_PRECAUTIOUS_BINARIES_SAVE
		}
#endif //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT
	}
	// write file with binaries

	return CL_SUCCESS;
}

int CCLProgramCompiler::Get_BuildLog(std::string &r_s_build_log, cl_build_status &r_n_build_status,
	cl_program h_program, cl_device_id h_device)
{
	r_s_build_log.erase();
	r_n_build_status = CL_BUILD_NONE;
	// clear output

	for(;;) {
		cl_int n_result;
		if((n_result = clGetProgramBuildInfo(h_program, h_device, CL_PROGRAM_BUILD_STATUS,
		   sizeof(cl_build_status), &r_n_build_status, NULL)) != CL_SUCCESS)
			return n_result;
		if(r_n_build_status != CL_BUILD_IN_PROGRESS)
			break;
#if defined(_WIN32) || defined(_WIN64)
		Sleep(100);
#else //_WIN32 || _WIN64
		sleep(1);
#endif //_WIN32 || _WIN64
	}
	// wait while build is in progress

	if(r_n_build_status == CL_BUILD_NONE)
		return CL_SUCCESS;
	// check build status

	size_t n_build_log_size;
	{
		cl_int n_result;
		if((n_result = clGetProgramBuildInfo(h_program, h_device,
		   CL_PROGRAM_BUILD_LOG, 0, NULL, &n_build_log_size)) != CL_SUCCESS)
			return n_result;
	}
	try {
		_ASSERTE(n_build_log_size != SIZE_MAX);
		r_s_build_log.resize(n_build_log_size + 1);
	} catch(std::bad_alloc&) {
		fprintf(stderr, "error: not enough memory for build log\n");
		return CL_OUT_OF_HOST_MEMORY;
	}
	// get build log length, allocate string buffer

	{
		cl_int n_result;
		if((n_result = clGetProgramBuildInfo(h_program, h_device, CL_PROGRAM_BUILD_LOG,
		   n_build_log_size, &r_s_build_log[0], NULL)) != CL_SUCCESS)
			return n_result;
	}
	r_s_build_log.resize(strlen(r_s_build_log.c_str()));
	// get build log, erase terminating null from string

	return CL_SUCCESS;
}

int CCLProgramCompiler::n_CompileProgramFile(cl_context h_context, cl_program *p_program,
	const char *p_s_source_file, size_t n_device_num, const cl_device_id *p_device_list,
	const char *p_s_build_options, const char *p_s_cache_file, int *p_status_word,
	size_t n_max_program_instance_num)
{
	int n_dummy_status_word;
	if(!p_status_word)
		p_status_word = &n_dummy_status_word;
	// makes sure p_status_word is not null

	*p_status_word = 0;
	// in case i/o errors occur below

	std::string s_program;
	{
		FILE *p_fr;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
		if(fopen_s(&p_fr, p_s_source_file, "rb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		if(!(p_fr = fopen(p_s_source_file, "rb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
			return CL_BUILD_PROGRAM_FAILURE; // not really great, but there's no better error code
		fseek(p_fr, 0, SEEK_END);
		size_t n_file_size = ftell(p_fr);
		try {
			s_program.resize(n_file_size);
		} catch(std::bad_alloc&) {
			fclose(p_fr);
			return CL_OUT_OF_HOST_MEMORY;
		}
		fseek(p_fr, 0, SEEK_SET);
		if(fread(&s_program[0], 1, n_file_size, p_fr) != n_file_size) {
			fclose(p_fr);
			return CL_BUILD_PROGRAM_FAILURE; // not really great, but there's no better error code
		}
		fclose(p_fr);
	}
	// read program from a file

	return n_CompileProgram(h_context, p_program, s_program.c_str(), n_device_num,
		p_device_list, p_s_build_options, p_s_cache_file, p_status_word, n_max_program_instance_num);
	// use the other function
}

#ifdef CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT

TSHA1 CCLProgramCompiler::t_Hash_ProgramSource_BuildOptions(const char *p_s_source, const char *p_s_build_options)
{
	CStreamHash<TSHA1> hash;
	hash.Process_Data(p_s_source, strlen(p_s_source) * sizeof(char));
	if(p_s_build_options && strcmp(p_s_build_options, ""))
		hash.Process_Data(p_s_build_options, strlen(p_s_build_options) * sizeof(char));
	return hash.t_Result();
}

int CCLProgramCompiler::n_ReadProgramBinaries(const char *p_s_filename, TSHA1 t_hash, cl_context h_context,
	cl_program *p_program, size_t n_device_num, const cl_device_id *p_device_list)
{
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb")))
		return cache_ReadFailed_FileNotFound;
	// open the file

	TSHA1 t_ref_hash;
	if(fread(&t_ref_hash[0], sizeof(uint32_t), 5, p_fr) != 5) {
		fclose(p_fr);
		return cache_ReadFailed_IO;
	}
	// read reference hash

	if(memcmp(&t_hash[0], &t_ref_hash[0], 5 * sizeof(uint32_t))) {
		fclose(p_fr);
		return cache_ReadFailed_SourceChecksum;
	}
	// compare it to current hash

	size_t n_binary_num;
	std::vector<size_t> binary_size_list;
	{
		uint32_t n_binary_num32;
		if(fread(&n_binary_num32, sizeof(uint32_t), 1, p_fr) != 1 || n_binary_num32 > SIZE_MAX) {
			fclose(p_fr);
			return cache_ReadFailed_IO;
		}
		n_binary_num = size_t(n_binary_num32);
		try {
			binary_size_list.resize(n_binary_num);
		} catch(std::bad_alloc&) {
			fclose(p_fr);
			return cache_ReadFailed_OutOfMemory;
		}
		for(size_t i = 0; i < n_binary_num; ++ i) {
			uint32_t n_binary_size32;
			if(fread(&n_binary_size32, sizeof(uint32_t), 1, p_fr) != 1 || n_binary_size32 > SIZE_MAX) {
				fclose(p_fr);
				return cache_ReadFailed_IO;
			}
			binary_size_list[i] = size_t(n_binary_size32);
		}
	}
	// read binary sizes

	size_t n_binaries_size = 0;
	for(size_t i = 0; i < n_binary_num; ++ i) {
		_ASSERTE(n_binaries_size <= SIZE_MAX - binary_size_list[i]);
		n_binaries_size += binary_size_list[i];
	}
	// sum binary sizes up

	TSHA1 t_binaries_hash;
	std::vector<const unsigned char*> binary_ptr_list;
	unsigned char *p_binaries;
	{
		try {
			binary_ptr_list.resize(n_binary_num);
			p_binaries = new unsigned char[n_binaries_size];
		} catch(std::bad_alloc&) {
			fclose(p_fr);
			return cache_ReadFailed_OutOfMemory;
		}
		// alloc buffers

		CStreamHash<TSHA1> hash;

		unsigned char *p_binary_ptr = p_binaries;
		for(size_t i = 0; i < n_binary_num; ++ i) {
			binary_ptr_list[i] = p_binary_ptr;

			size_t n_binary_size = binary_size_list[i];

			if(fread(p_binary_ptr, n_binary_size, 1, p_fr) != 1) {
				fclose(p_fr);
				return cache_ReadFailed_IO;
			}

			hash.Process_Data(p_binary_ptr, n_binary_size);

			p_binary_ptr += n_binary_size;
		}
		// get pointers, read binaries (so read block size is 32-bit), calculate their hashes

		t_binaries_hash = hash.t_Result();
	}
	// read binaries

	TSHA1 t_binaries_hash_ref;
	if(fread(&t_binaries_hash_ref[0], sizeof(uint32_t), 5, p_fr) != 5) {
		fclose(p_fr);
		return cache_ReadFailed_IO;
	}
	// read checksum

	fclose(p_fr);
	// close file

	if(memcmp(&t_binaries_hash_ref[0], &t_binaries_hash_ref[0], 5 * sizeof(uint32_t)))
		return cache_ReadFailed_BinaryChecksum;
	// make sure checksum is correct

#if defined CL_COMPILER_UTIL_BOGUS_DATA_DETECTION && defined(__DIR_INCLUDED)
	TFileInfo t_file(p_s_filename);
	if(t_file.n_Size64() != 2 * 5 * sizeof(uint32_t) + sizeof(uint32_t) + // size of two hashes, number of binaries
	   n_binary_num * sizeof(uint32_t) + n_binaries_size) // lengths of binaries, binaries data
		return cache_ReadFailed_BinaryChecksum; // ...
	// make sure file doesn't contain bogus data (paranoid)
#endif // CL_COMPILER_UTIL_BOGUS_DATA_DETECTION && __DIR_INCLUDED

	cl_int n_result;
	cl_program h_prog = clCreateProgramWithBinary(h_context, n_device_num, p_device_list,
		&binary_size_list[0], &binary_ptr_list[0], NULL, &n_result);
	if(n_result != CL_SUCCESS)
		return cache_ReadFailed_CreateProgram;
	if((n_result = clBuildProgram(h_prog, 0, NULL, NULL, NULL, NULL)) != CL_SUCCESS)
		return cache_ReadFailed_CreateProgram;
	// use loaded binaries to create the program ...

	*p_program = h_prog;
	// write output

	return cache_ReadSucceeded;
}

bool CCLProgramCompiler::WriteProgramBinaries(cl_program h_program, TSHA1 t_hash, const char *p_s_filename)
{
	cl_uint n_device_num;
	if(clGetProgramInfo(h_program, CL_PROGRAM_NUM_DEVICES,
	   sizeof(size_t), &n_device_num, NULL) != CL_SUCCESS)
		return false;
	// get number of devices

	if(!n_device_num || n_device_num > UINT32_MAX)
		return false;
	// we can't query binaries without devices

	size_t n_binary_num = n_device_num;

	std::vector<size_t> binary_size_list;
	try {
#ifdef _DEBUG
		_ASSERTE(n_binary_num < SIZE_MAX);
		binary_size_list.resize(n_binary_num + 1);
		binary_size_list[n_binary_num] = 0xbaadf00d; // "magic" word
#else //_DEBUG
		binary_size_list.resize(n_binary_num);
#endif //_DEBUG
	} catch(std::bad_alloc&) {
		return false;
	}
	if(clGetProgramInfo(h_program, CL_PROGRAM_BINARY_SIZES,
	   n_binary_num * sizeof(size_t), &binary_size_list[0], NULL) != CL_SUCCESS)
		return false;
#ifdef _DEBUG
	_ASSERTE(binary_size_list[n_binary_num] == 0xbaadf00d); // make sure it's returning the right amount of data
#endif //_DEBUG
	// get binary sizes

	size_t n_binaries_size = 0;
	for(size_t i = 0; i < n_binary_num; ++ i) {
		_ASSERTE(n_binaries_size <= SIZE_MAX - binary_size_list[i]);
		n_binaries_size += binary_size_list[i];
	}
	// sum binary sizes up

	FILE *p_fw;
	if(!(p_fw = fopen(p_s_filename, "wb")))
		return false;
	// open file for writing ...

	if(fwrite(&t_hash[0], sizeof(uint32_t), 5, p_fw) != 5) {
		fclose(p_fw);
		return false;
	}
	// write hash

	uint32_t n_binary_num32 = uint32_t(n_binary_num);
	if(fwrite(&n_binary_num32, sizeof(uint32_t), 1, p_fw) != 1) {
		fclose(p_fw);
		return false;
	}
	// write number of binaries

	for(size_t i = 0; i < n_binary_num; ++ i) {
		_ASSERTE(binary_size_list[i] <= UINT32_MAX);
		uint32_t n_binary_size32 = binary_size_list[i];
		if(fwrite(&n_binary_size32, sizeof(uint32_t), 1, p_fw) != 1) {
			fclose(p_fw);
			return false;
		}
	}
	// write size of each binary

	unsigned char *p_binaries;
	std::vector<const unsigned char*> binary_ptr_list;
	{
		try {
			binary_ptr_list.resize(n_binary_num);
			p_binaries = new unsigned char[n_binaries_size];
		} catch(std::bad_alloc&) {
			fclose(p_fw);
			return false;
		}
		// alloc buffers

		unsigned char *p_binary_ptr = p_binaries;
		for(size_t i = 0; i < n_binary_num; ++ i) {
			binary_ptr_list[i] = p_binary_ptr;
			p_binary_ptr += binary_size_list[i];
		}
		// get pointers
	}
	if(clGetProgramInfo(h_program, CL_PROGRAM_BINARIES,
	   n_binaries_size, &binary_ptr_list[0], NULL) != CL_SUCCESS) {
		fclose(p_fw);
		delete[] p_binaries;
		return false;
	}
	// get binaries

	TSHA1 t_binaries_hash;
	{
		CStreamHash<TSHA1> hash;
		if(fwrite(p_binaries, n_binaries_size, 1, p_fw) != 1) {
			fclose(p_fw);
			delete[] p_binaries;
			return false;
		}
		hash.Process_Data(p_binaries, n_binaries_size);
		t_binaries_hash = hash.t_Result();
	}
	// write binaries, calculate hash

	delete[] p_binaries;
	// cleanup

	if(fwrite(&t_binaries_hash[0], sizeof(uint32_t), 5, p_fw) != 5) {
		fclose(p_fw);
		return false;
	}
	// write hash

	fclose(p_fw);
	// close output file

	return true;
}

#endif //CL_PROGRAM_COMPILER_USE_OLD_BINARIES_FORMAT

/*
 *								=== ~CCLProgramCompiler ===
 */

/*
 *								=== CCLArgLoaderHelper ===
 */

int CCLArgLoaderHelper::__SafeCall(int n_error_code, const char *p_s_file, int n_line)
{
	if(n_error_code != CL_SUCCESS) {
		fprintf(stderr, "error: clParamSet*() failed : error code %04d : "
			"file \'%s\', line %d.\n", n_error_code, p_s_file, n_line);
		exit(-1);
	}

	return n_error_code;
}

/*
 *								=== ~CCLArgLoaderHelper ===
 */
