/**
 *	@file Main.cpp
 *	@date 2010-12-05
 *	@brief A simple OpenCL demonstration program (calculates vector addition on very long arrays).
 *	@author -tHE SWINe-
 *
 *	This demonstrates basic use of CCLUtils, CCLDeviceParams, CCLProgramCompiler and clSetKernelArgs().
 *	Note this source is about 40 lines shorter, compared to NVIDIA oclVectorAdd sdk sample.
 */

#include <stdio.h>
#include <math.h>
#include <CL/opencl.h>
#include "ClUtils.h"
#include "Timer.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// MSVC 'for' scoping hack

int main(int n_arg_num, const char **p_arg_list)
{
	bool b_verbose = true;
	size_t n_vector_length = 1 << 26;
	// input parameters

	int *p_vector_a, *p_vector_b, *p_vector_c, *p_gpu_result;
	if(!(p_vector_a = new(std::nothrow) int[n_vector_length]) ||
	   !(p_vector_b = new(std::nothrow) int[n_vector_length]) ||
	   !(p_vector_c = new(std::nothrow) int[n_vector_length]) ||
	   !(p_gpu_result = new(std::nothrow) int[n_vector_length])) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// allocate buffers

	for(size_t i = 0; i < n_vector_length; ++ i) {
		p_vector_a[i] = rand();
		p_vector_b[i] = rand();
	}
	// generate random input data

	CTimer timer;
	timer.ResetTimer();

	double f_cpu_start = timer.f_Time();

	{
		for(size_t i = 0; i < n_vector_length; ++ i)
			p_vector_c[i] = p_vector_a[i] + p_vector_b[i];
		// run the calculation on CPU
	}
	// CPU version of the algorithm

	double f_cpu_time = timer.f_Time() - f_cpu_start;

	double f_gpu_time;

	{
		if(b_verbose)
			printf("initializing OpenCL ...\n");

		cl_context h_context;
		{
			int n_result;
			if((n_result = CCLUtils::n_OpenCL_Init(&h_context)) != CL_SUCCESS) {
				fprintf(stderr, "error: failed to initialize OpenCL (%d)\n", n_result);
				return -1;
			}
		}
		// create OpenCL context

		cl_device_id h_device;
		if(CCLUtils::n_Get_MaxGFlops_DeviceId(&h_device, h_context) < 0) {
			fprintf(stderr, "error: failed to get handle of the fastest OpenCL device\n");
			return -1;
		}
		// get fastest device

		CCLDeviceParams device_params(h_device);
		if(b_verbose)
			device_params.Dump(); // @note CCLDeviceParams actually has some useful functions
		// show some info about the device

		cl_command_queue h_cmd_queue;
		{
			cl_int n_result;
			h_cmd_queue = clCreateCommandQueue(h_context, h_device, CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &n_result);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to create OpenCL command queue (%d)\n", n_result);
				return -1;
			}
		}
		// create command queue

		if(b_verbose)
			printf("loading \'%s\' ... ", "Kernel.c");
		// verbose

		cl_program h_program;
		{
			char p_s_preprocessor[256];
			sprintf(p_s_preprocessor, "-D VECTOR_LENGTH=%d", n_vector_length); // this is mostly useless in this scenario ... but comes in handy with more complex kernels
			// create preprocessor string, based on commandline params

			int n_flags;
			int n_result = CCLProgramCompiler::n_CompileProgramFile(h_context, &h_program, "Kernel.c",
				1, &h_device, p_s_preprocessor, "Kernel.clbin", &n_flags);
			// compile the program

			if(b_verbose) 
				CCLProgramCompiler::Dump_StatusWord(n_flags);
			// verbose (see what steps did it take)

			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to load OpenCL program (%d)\n", n_result);
				return -1;
			}
		}
		// compile program

		cl_kernel h_vec_add_kernel;
		{
			cl_int n_result;
			h_vec_add_kernel = clCreateKernel(h_program, "VectorAdd", &n_result);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to create OpenCL kernel (%d)\n", n_result);
				return -1;
			}
		}
		// get OpenCL kernel

		cl_mem dp_vector_a, dp_vector_b, dp_vector_c;
		{
			cl_int n_result;
			dp_vector_a = clCreateBuffer(h_context, CL_MEM_READ_ONLY, n_vector_length * sizeof(int), NULL, &n_result);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to alloc device buffer (%d)\n", n_result);
				return -1;
			}
			dp_vector_b = clCreateBuffer(h_context, CL_MEM_READ_ONLY, n_vector_length * sizeof(int), NULL, &n_result);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to alloc device buffer (%d)\n", n_result);
				return -1;
			}
			dp_vector_c = clCreateBuffer(h_context, CL_MEM_WRITE_ONLY, n_vector_length * sizeof(int), NULL, &n_result);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to alloc device buffer (%d)\n", n_result);
				return -1;
			}
		}
		// alloc OpenCL buffers

		{
			int n_result;
			if((n_result = clEnqueueWriteBuffer(h_cmd_queue, dp_vector_a, false, 0, n_vector_length * sizeof(int),
			   p_vector_a, 0, NULL, NULL)) != CL_SUCCESS) {
				fprintf(stderr, "error: failed to copy data to the device (%d)\n", n_result);
				return -1;
			}
			if((n_result = clEnqueueWriteBuffer(h_cmd_queue, dp_vector_b, false, 0, n_vector_length * sizeof(int),
			   p_vector_b, 0, NULL, NULL)) != CL_SUCCESS) {
				fprintf(stderr, "error: failed to copy data to the device (%d)\n", n_result);
				return -1;
			}
		}
		// copy data from host to device (non-blocking calls)

		{
			int n_result;
			if((n_result = clFinish(h_cmd_queue)) != CL_SUCCESS) {
				fprintf(stderr, "error: clFinish() failed (%d)\n", n_result);
				return -1;
			}
		}
		// wait for GPU to finish

		double f_gpu_start = timer.f_Time();

		{
			size_t n_local_work_size = std::min(256, device_params.t_Properties().maxThreadsPerBlock); // prefer to use 256 threads / block
			size_t n_global_work_size = n_Align_Up(n_vector_length, n_local_work_size);
			// calculate local and global work size

			clSetKernelArgs(h_vec_add_kernel, dp_vector_c, dp_vector_a, dp_vector_b);
			// set kernel parameters (a ClUtils function)

			int n_result = clEnqueueNDRangeKernel(h_cmd_queue, h_vec_add_kernel, 1,
				NULL, &n_global_work_size, &n_local_work_size, 0, NULL, NULL);
			if(n_result != CL_SUCCESS) {
				fprintf(stderr, "error: failed to launch OpenCL kernel (gs %d, ls %d, result %d)\n",
					n_global_work_size, n_local_work_size, n_result);
				return -1;
			}
			// launch kernel
		}

		{
			int n_result;
			if((n_result = clFinish(h_cmd_queue)) != CL_SUCCESS) {
				fprintf(stderr, "error: clFinish() failed (%d)\n", n_result);
				return -1;
			}
		}
		// wait for GPU to finish

		f_gpu_time = timer.f_Time() - f_gpu_start;

		if(clEnqueueReadBuffer(h_cmd_queue, dp_vector_c, true, 0, n_vector_length * sizeof(int),
		   p_gpu_result, 0, NULL, NULL) != CL_SUCCESS) {
			fprintf(stderr, "error: failed to copy data from the device\n");
			return -1;
		}
		// copy data back to system memory

		if(!memcmp(p_gpu_result, p_vector_c, n_vector_length * sizeof(int)))
			printf("ok, GPU results match CPU\n");
		else
			fprintf(stderr, "error: results calcualted on GPU do not match CPU results\n");
		// check results

		clReleaseKernel(h_vec_add_kernel);
		clReleaseMemObject(dp_vector_a);
		clReleaseMemObject(dp_vector_b);
		clReleaseMemObject(dp_vector_c);
		clReleaseProgram(h_program);
		clReleaseCommandQueue(h_cmd_queue);
		clReleaseContext(h_context);
		// cleanup

		if(b_verbose)
			printf("finished ...\n");
		// verbose
	}
	// GPU version of the algorithm

	printf("CPU time: %f\n", f_cpu_time);
	printf("GPU time: %f\n", f_gpu_time);

	delete[] p_vector_a;
	delete[] p_vector_b;
	delete[] p_vector_c;
	delete[] p_gpu_result;

	return 0;
}
